from __future__ import division, print_function, absolute_import, unicode_literals

import os
import hashlib
from mog_commons.case_class import CaseClass
from mog_commons.collection import get_single_item
from mog_commons.functional import oget
from mog_commons.types import *
from mog_commons.string import to_unicode, is_strlike, to_bytes
from easy_menu.entity import Meta


class CommandLine(CaseClass):
    @types(cmd=Unicode, meta=Meta, encoding=String)
    def __init__(self, cmd, meta, encoding='utf-8'):
        """
        :param cmd: command line string
        :param meta:
        :param encoding: encoding for command line string
        :return:
        """
        CaseClass.__init__(self, ('cmd', cmd), ('meta', meta), ('encoding', encoding))

    @staticmethod
    def parse(data, meta, encoding='utf-8'):
        """
        Parse one command line.
        :param data: string or dict:
        :param meta: Meta: meta configuration inherited from the parent menu
        :param encoding: string:
        :return: CommandLine:
        """
        assert isinstance(meta, Meta)

        def f(s):
            return to_unicode(s, encoding)

        if is_strlike(data):
            return CommandLine(f(data), meta, encoding)
        elif isinstance(data, dict):
            cmd, params = get_single_item(data)
            assert is_strlike(cmd), 'cmd must be string, not %s.' % type(cmd).__name__
            new_meta = meta.updated(params, encoding)
            return CommandLine(to_unicode(cmd, encoding), new_meta, encoding)
        else:
            raise ValueError('CommandLine must be string or dict, not %s.' % type(data).__name__)

    @types(Unicode)
    def formatted(self):
        buf = [
            '- cmd: %s' % self.cmd,
            '  cwd: %s' % oget(self.meta.work_dir, os.path.abspath(os.path.curdir)),
        ]
        if self.meta.env:
            buf.append('  env: {%s}' % ', '.join('%s: %s' % (k, v) for k, v in sorted(self.meta.env.items())))
        if self.meta.lock:
            buf.append('  lock: True')
        return '\n'.join(buf)

    @types(String)
    def to_hash_string(self):
        # ignore work directory
        data = b''.join(to_bytes(s) for s in [self.cmd, sorted(self.meta.env.items())])

        return hashlib.md5(data).hexdigest()



from __future__ import division, print_function, absolute_import, unicode_literals

import six
from mog_commons.string import to_unicode
from mog_commons.collection import get_single_item
from mog_commons.types import *
from easy_menu.entity import Meta, Item


class Menu(Item):
    """
    Menu is built from an one-element dict having title string as key and item list element as value
    """

    @types(title=Unicode, items=ListOf(Item), meta=Meta)
    def __init__(self, title, items, meta=Meta()):
        """
        :param title:
        :param items:
        :param meta:
        :return:
        """
        Item.__init__(self, ('title', title), ('items', items), ('meta', meta))

    @staticmethod
    @types(Item, data=dict, meta=Meta)
    def parse(data, meta, loader, encoding='utf-8', depth=0):
        """
        :param data:
        :param meta:
        :param loader:
        :param encoding:
        :param depth:
        :return:
        """
        from easy_menu.entity import KEYWORD_META, Item

        # read meta configurations
        if KEYWORD_META in data:
            meta = meta.updated(data[KEYWORD_META], encoding)
            del data[KEYWORD_META]

        assert len(data) == 1, 'Menu should have only one item, not %s.' % len(data)

        title, content = get_single_item(data)
        assert isinstance(title, six.string_types), 'Menu title must be string, not %s.' % type(title).__name__
        assert isinstance(content, list), 'Menu content must be list, not %s.' % type(content).__name__
        title = to_unicode(title, encoding)

        items = [Item.parse(item, meta, loader, encoding, depth + 1) for item in content]
        return Menu(title, items, meta)

    @types(Unicode)
    def formatted(self):
        """Return formatted string for pretty printing."""
        return '\n'.join(
            ['# %s:' % self.title] + ['  %s' % line for x in self.items for line in x.formatted().splitlines()])



from __future__ import division, print_function, absolute_import, unicode_literals

import os
import sys
import locale
from mog_commons.case_class import CaseClass
from mog_commons.functional import omap, oget
from mog_commons.string import to_unicode

from easy_menu.setting import arg_parser
from easy_menu.setting.loader import Loader
from easy_menu.entity import Menu, Meta
from easy_menu.exceptions import SettingError, ConfigError

DEFAULT_CONFIG_NAME = os.environ.get('EASY_MENU_CONFIG', 'easy-menu.yml')
EVAL_CACHE_DIR = os.path.join(os.path.expanduser('~'), '.easy-menu', 'eval')
COMMAND_PID_DIR = os.path.join(os.path.expanduser('~'), '.easy-menu', 'pid')


class Setting(CaseClass):
    """
    Manages all settings.
    """

    def __init__(self, config_path=None, work_dir=None, root_menu=None, encoding=None, lang=None, width=None,
                 clear_cache=False, cache_dir=EVAL_CACHE_DIR, pid_dir=COMMAND_PID_DIR,
                 stdin=None, stdout=None, stderr=None, getch_enabled=True, source_enabled=True):
        is_url = Loader.is_url(config_path)
        work_dir = omap(lambda s: to_unicode(s, encoding), self._search_work_dir(work_dir, config_path, is_url))

        CaseClass.__init__(self,
                           ('config_path', config_path),
                           ('work_dir', work_dir),
                           ('root_menu', oget(root_menu, {})),
                           ('encoding', encoding),
                           ('lang', self._find_lang(lang)),
                           ('width', width),
                           ('clear_cache', clear_cache),
                           ('cache_dir', cache_dir),
                           ('pid_dir', pid_dir),
                           ('stdin', oget(stdin, sys.stdin)),
                           ('stdout', oget(stdout, sys.stdout)),
                           ('stderr', oget(stderr, sys.stderr)),
                           ('getch_enabled', getch_enabled),
                           ('source_enabled', source_enabled)
                           )

    @staticmethod
    def _find_lang(lang):
        if not lang:
            # environment LANG is the first priority
            lang = os.environ.get('LANG')
        if not lang:
            lang = locale.getdefaultlocale()[0]
        return lang

    @staticmethod
    def _search_work_dir(work_dir, config_path, is_url):
        if work_dir is None:
            if config_path is not None:
                if not is_url:
                    return os.path.dirname(config_path)
        return work_dir

    def resolve_encoding(self, terminal_handler):
        if not self.encoding:
            return self.copy(encoding=terminal_handler.encoding)
        return self

    def parse_args(self, argv):
        option, args = arg_parser.parser.parse_args(argv[1:])
        path = None

        if not args:
            pass
        elif len(args) == 1:
            path = args[0]
            if not Loader.is_url(path):
                path = os.path.abspath(path)
        else:
            arg_parser.parser.print_help()
            arg_parser.parser.exit(2)

        return self.copy(config_path=path, work_dir=option.work_dir, encoding=option.encoding, lang=option.lang,
                         width=option.width, clear_cache=option.clear_cache, getch_enabled=option.getch_enabled)

    def lookup_config(self):
        if self.config_path is None:
            d = os.path.abspath(self.work_dir) if self.work_dir else os.getcwd()

            while True:
                path = os.path.join(d, DEFAULT_CONFIG_NAME)
                if os.path.exists(path):
                    return self.copy(config_path=path)
                nd = os.path.dirname(d)
                if d == nd:
                    break
                d = nd
        return self

    def load_config(self):
        """
        Load the configuration file or url.

        If it contains 'include' sections, load them recursively.
        :return: updated Setting instance
        """
        if self.config_path is None:
            raise SettingError('Not found configuration file.')

        loader = Loader(self.work_dir, self.cache_dir, self.encoding, self.stdout, self.clear_cache)
        data = loader.load(False, self.config_path)
        try:
            root_menu = Menu.parse(data, Meta(self.work_dir), loader, self.encoding, 0)
        except (AssertionError, ValueError, TypeError) as e:
            raise ConfigError(self.config_path, e)
        return self.copy(root_menu=root_menu)



from __future__ import division, print_function, absolute_import, unicode_literals

from abc import ABCMeta, abstractmethod
import six
from mog_commons.case_class import CaseClass
from mog_commons.collection import get_single_item
from mog_commons.types import *
from easy_menu.entity.meta import Meta
from easy_menu.setting.loader import Loader


@six.add_metaclass(ABCMeta)
class Item(CaseClass):
    """
    Abstract item class
    """

    @staticmethod
    def _is_command_like(content):
        """
        We assume the data is command when it is a string or one-element dict which contains string key and dict value
        :param content: object to parse
        :return: true when the data is command-like
        """
        if isinstance(content, six.string_types):
            return True

        if isinstance(content, list):
            for item in content:
                ok = False
                if isinstance(item, six.string_types):
                    ok = True
                if isinstance(item, dict) and len(item) == 1:
                    cmd, attr = get_single_item(item)
                    ok = isinstance(cmd, six.string_types) and isinstance(attr, dict)
                if not ok:
                    return False
            return True

        return False

    @staticmethod
    @types(data=dict, meta=Meta, loader=Loader)
    def parse(data, meta, loader, encoding='utf-8', depth=0):
        """
        :param data:
        :param meta:
        :param loader:
        :param encoding:
        :param depth: indicator for the nesting level
        :return:
        """
        from easy_menu.entity import Menu, Command, KEYWORD_META, KEYWORD_INCLUDE, KEYWORD_EVAL

        # avoid for inclusion loops and stack overflow
        assert depth < 50, 'Nesting level too deep.'

        # if the data has meta key, it should be a menu.
        if KEYWORD_META in data:
            return Menu.parse(data, meta, loader, encoding, depth)

        # parse eval cache setting
        eval_expire = None
        if KEYWORD_EVAL in data:
            if 'cache' in data:
                eval_expire = data['cache']
                del data['cache']

        assert len(data) == 1, 'Item should have only one element, not %s.' % len(data)

        title, content = get_single_item(data)

        if title == KEYWORD_INCLUDE:
            assert isinstance(content, six.string_types), \
                '"include" section must have string content, not %s.' % type(content).__name__
            return Menu.parse(loader.load(False, content), meta, loader, encoding, depth)
        elif title == KEYWORD_EVAL:
            assert isinstance(content, six.string_types), \
                '"eval" section must have string content, not %s.' % type(content).__name__
            return Menu.parse(loader.load(True, content, eval_expire), meta, loader, encoding, depth)
        elif Item._is_command_like(content):
            return Command.parse(data, meta, loader, encoding, depth)
        else:
            return Menu.parse(data, meta, loader, encoding, depth)

    @abstractmethod
    def formatted(self):
        """abstract method"""



from __future__ import division, print_function, absolute_import, unicode_literals

import copy
from mog_commons.case_class import CaseClass
from mog_commons.types import *
from mog_commons.string import to_unicode


class Meta(CaseClass):
    """
    Meta settings for running commands
    """

    @types(work_dir=Option(Unicode), env=Option(DictOf(Unicode, Unicode)), lock=bool)
    def __init__(self, work_dir=None, env=None, lock=False):
        """
        :param work_dir:
        :param env:
        :param lock:
        :return:
        """
        env = env or {}
        CaseClass.__init__(self, ('work_dir', work_dir), ('env', env), ('lock', lock))

    @types(data=dict)
    def updated(self, data, encoding):
        """
        Load configuration and return updated instance.
        :param data: dict:
        :return: Meta:
        """
        functions = {
            'work_dir': Meta._load_work_dir,
            'env': Meta._load_env,
            'lock': Meta._load_lock,
        }

        ret = self.copy()
        for k, v in data.items():
            ret = functions.get(k, self.__unknown_field(k))(ret, v, encoding)
        return ret

    @types(data=String, encoding=String)
    def _load_work_dir(self, data, encoding):
        """Overwrite working directory"""
        self.work_dir = to_unicode(data, encoding)
        return self

    @types(data=DictOf(String, String), encoding=String)
    def _load_env(self, data, encoding):
        """Merge environment variables"""
        d = copy.copy(self.env)
        d.update([(to_unicode(k, encoding), to_unicode(v, encoding)) for k, v in data.items()])
        self.env = d
        return self

    @types(data=bool, encoding=String)
    def _load_lock(self, data, encoding):
        """Overwrite lock setting"""
        self.lock = data
        return self

    @staticmethod
    def __unknown_field(key):
        def f(x, y, z):
            raise ValueError('Unknown field: %s' % key)

        return f



from __future__ import division, print_function, absolute_import, unicode_literals

import sys
from datetime import datetime

from mog_commons.string import *
from mog_commons.io import print_safe
from mog_commons.types import *
from easy_menu.entity import Menu, Command
from easy_menu.exceptions import EncodingError, SettingError
from easy_menu.view import i18n

DEFAULT_WINDOW_WIDTH = 78
DEFAULT_PAGE_SIZE = 9


class Terminal(object):
    def __init__(self, root_menu, host, user, executor, handler, width=None, page_size=None, _input=sys.stdin,
                 _output=sys.stdout, encoding=None, lang=None, timing=True, source_enabled=True):
        """
        :param root_menu: dict of root menu
        :param host: host name string
        :param user: user name string
        :param executor: Executor instance
        :param handler: mog_common.terminal.TerminalHandler instance
        :param width:
        :param page_size:
        :param _input:
        :param _output:
        :param encoding:
        :param lang: language setting
        :param timing: bool: print running time after executing command if true
        :param source_enabled: bool: allow source printing if true
        :return:
        """
        # fields
        self.root_menu = root_menu
        self.host = host
        self.user = user
        self.executor = executor
        self.handler = handler
        self.width = DEFAULT_WINDOW_WIDTH if width is None else width
        self.page_size = DEFAULT_PAGE_SIZE if page_size is None else page_size
        self._input = _input
        self._output = _output
        self.encoding = encoding
        self.lang = lang
        self.i18n = self._find_i18n(lang)
        self.timing = timing
        self.source_enabled = source_enabled

        if self.width < 40:
            raise SettingError('width must be equal or greater than 40: width=%s' % self.width)
        if not 0 < self.page_size <= 9:
            raise SettingError('page_size must be positive and one digit: page_size=%s' % self.page_size)

    @staticmethod
    def _find_i18n(lang):
        if not lang:
            return i18n.messages_en
        elif lang.lower().startswith('ja'):
            return i18n.messages_ja
        else:
            return i18n.messages_en

    def _num_pages(self, num_items):
        return max(1, (num_items + self.page_size - 1) // self.page_size)

    #
    # Design parts
    #
    def thin_line(self):
        return '-' * self.width

    def thick_line(self):
        return '=' * self.width

    def header_line(self):
        return edge_just(self.i18n.MSG_HOST % self.host, self.i18n.MSG_USER % self.user, self.width)

    @staticmethod
    def title_line(title):
        return ' ' * 2 + title

    def menu_line(self):
        return '+'.rjust(7, '-').ljust(self.width, '-')

    def pager_line(self, offset, num_pages):
        left = '<= [P]' if 0 < offset else '      '
        right = '[N] =>' if offset + 1 < num_pages else '      '
        middle = ('Page %d / %d' % (offset + 1, num_pages)).center(self.width - len(left) - len(right))
        return left + middle + right

    #
    # Building output string
    #
    def _get_header(self, title):
        return [
            self.header_line(),
            self.thick_line(),
            self.title_line(title),
            self.thin_line(),
        ]

    def _get_footer(self, message):
        return [
            self.thick_line(),
            message,
        ]

    def _get_description(self, item):
        return (self.i18n.MSG_SUB_MENU if isinstance(item, Menu) else '%s') % item.title

    def _get_breadcrumb(self, titles):
        s = ' > '.join(titles)
        n = unicode_width(s)
        limit = self.width - 5
        return ('~' if limit < n else '') + unicode_right(s, limit)

    def get_page(self, titles, page_items, offset, num_pages):
        """Make menu page string."""

        assert len(page_items) <= self.page_size, 'Number of page items must less or equal than page size.'

        title = self._get_breadcrumb(titles)

        pager_lines = [] if num_pages <= 1 else [
            self.pager_line(offset, num_pages),
            self.thin_line(),
        ]

        item_lines = [self.i18n.MSG_ITEM % (i + 1, self._get_description(item)) for i, item in enumerate(page_items)]

        quit_lines = [
            self.menu_line(),
            self.i18n.MSG_ITEM % (
                0, self.i18n.MSG_QUIT if len(titles) == 1 else self.i18n.MSG_RETURN % titles[-2]),
        ]

        message = self.i18n.MSG_INPUT_NUM % (0, len(page_items))
        return '\n'.join(self._get_header(title) + pager_lines + item_lines + quit_lines + self._get_footer(message))

    def get_confirm(self, description):
        """
        Make confirmation page string

        :param description: description for the command to execute
        :return: string
        """

        item_lines = [
            self.i18n.MSG_CONFIRM % description
        ]
        return '\n'.join(
            self._get_header(self.i18n.MSG_CONFIRM_TITLE) + item_lines +
            self._get_footer(self.i18n.MSG_CONFIRM_QUESTION))

    def get_duplicate(self, description):
        """
        Make duplicate check page string
        """

        item_lines = [
            self.i18n.MSG_DUPLICATE % description
        ]
        return '\n'.join(
            self._get_header(self.i18n.MSG_DUPLICATE_TITLE) + item_lines +
            self._get_footer(self.i18n.MSG_DUPLICATE_QUESTION)
        )

    def get_before_execute(self, description):
        return '\n'.join(self._get_header(self.i18n.MSG_RUN_TITLE % description) + [''])

    def get_after_execute(self, title, return_code, start_time, end_time):
        items = [('Return code', '%d' % return_code)]
        if self.timing:
            items = [
                (self.i18n.MSG_FINISH, title),
                ('Running time', '%s  (%s -> %s)' % (self._format_timedelta(end_time - start_time),
                                                     self._format_datetime(start_time),
                                                     self._format_datetime(end_time))),
            ] + items
        key_width = max(unicode_width(x[0]) for x in items)
        result_lines = [self.thin_line()] + ['%s: %s' % (unicode_ljust(k, key_width), v) for k, v in items]
        return '\n'.join(result_lines + self._get_footer(self.i18n.MSG_INPUT_ANY))

    @staticmethod
    def _format_timedelta(tm):
        hour, second = divmod(tm.seconds, 60 * 60)
        minute, second = divmod(second, 60)

        if tm.days < 0:
            return ''
        elif tm.days > 0:
            return '%dd %dh %dm %ds' % (tm.days, hour, minute, second)
        elif hour > 0:
            return '%dh %dm %ds' % (hour, minute, second)
        elif minute > 0:
            return '%dm %ds' % (minute, second)
        elif second > 0:
            return '%ds' % second
        return '%dms' % (tm.microseconds // 1000)

    @staticmethod
    def _format_datetime(dt):
        return dt.strftime('%Y-%m-%d %H:%M:%S')

    #
    # Wait for input
    #
    def wait_input_char(self):
        ch = self.handler.getch()
        if ch in ['\x03', '\x04']:
            # pressed C-c or C-d
            raise KeyboardInterrupt
        return ch

    def wait_input_yes_no(self, default=False):
        while True:
            ch = self.wait_input_char().lower()
            if ch == 'y':
                return True
            elif ch == 'n':
                return False
            elif ch == '\r' or not self.handler.getch_enabled:
                # default
                return default

    #
    # Output
    #
    def _print(self, unicode_text):
        assert is_unicode(unicode_text), 'Text must be unicode: %s' % unicode_text

        try:
            print_safe(unicode_text, self.encoding, output=self._output, errors='strict', newline='')
        except (LookupError, UnicodeError):
            raise EncodingError('Failed to print menu: lang=%s, encoding=%s' % (self.lang, self.encoding))

    def _draw(self, unicode_text):
        self.handler.clear()
        self._print(unicode_text)

    #
    # Main loop
    #
    def wait_input_menu(self, menu_items, offset, num_pages):
        """
        :param menu_items: list of items
        :param offset: page offset
        :return: function(stack, offset => stack, offset)
        """
        num_items = len(menu_items)

        while True:
            ch = self.wait_input_char().lower()
            if ch == '0':
                return lambda s, o: (s[:-1], 0)
            elif ch.isdigit():
                index = offset * self.page_size + int(ch) - 1
                if index < num_items:
                    item = menu_items[index]

                    # check if it is a sub menu
                    if isinstance(item, Menu):
                        return lambda s, o: (s + [item], 0)

                    def f(s, o):  # side effect only
                        self.execute_command(item)
                        return s, o

                    return f
            elif ch == 'n' and offset + 1 < num_pages:
                return lambda s, o: (s, o + 1)
            elif ch == 'p' and 0 < offset:
                return lambda s, o: (s, o - 1)
            elif ch == 's' and self.source_enabled:
                def f(s, o):  # side effect only
                    self.print_source()
                    return s, o
                return f

            # if getch is disabled, redraw screen
            if not self.handler.getch_enabled:
                return lambda s, o: (s, o)

    @types(command=Command)
    def execute_command(self, command):
        """
        Confirm with prompt before executing command.

        :param command: Command:
        :return: None
        """
        # confirmation
        self._draw(self.get_confirm(command.title))
        if not self.wait_input_yes_no():
            return

        # duplicate check
        if self.executor.is_running(command):
            self._draw(self.get_duplicate(command.title))
            if not self.wait_input_yes_no():
                return

        # run command
        start_time = datetime.now()
        self._draw(self.get_before_execute(command.title))
        return_code = self.executor.execute(command)
        end_time = datetime.now()
        if return_code == 130:
            # maybe interrupted
            self._print('\n')

        self._print(self.get_after_execute(command.title, return_code, start_time, end_time))
        self.wait_input_char()  # wait for any input

    def print_source(self):
        """Print source of the root menu."""
        self._draw('\n'.join(self._get_header(self.i18n.MSG_SOURCE_TITLE)))
        self._print('\n%s\n' % self.root_menu.formatted())
        self._print('\n'.join(self._get_footer(self.i18n.MSG_INPUT_ANY)))
        self.wait_input_char()

    def loop(self):
        stack = [self.root_menu]
        offset = 0  # current page index

        while stack:
            titles = [menu.title for menu in stack]
            items = stack[-1].items
            num_pages = self._num_pages(len(items))

            # apply offset
            page_items = items[self.page_size * offset:self.page_size * (offset + 1)]

            self._draw(self.get_page(titles, page_items, offset, num_pages))

            f = self.wait_input_menu(items, offset, num_pages)
            stack, offset = f(stack, offset)

        self.handler.clear()



from __future__ import division, print_function, absolute_import, unicode_literals

from abc import ABCMeta, abstractmethod
import six


@six.add_metaclass(ABCMeta)
class Executor(object):
    """abstract executor class"""

    @abstractmethod
    def execute(self, command):
        """abstract method"""

    def is_running(self, command):
        """abstract method"""



from __future__ import division, print_function, absolute_import, unicode_literals

import os
import sys
import time
import re
import yaml
import hashlib
from six.moves.urllib.request import urlopen
from jinja2 import Environment
from mog_commons.command import capture_command
from mog_commons.string import *
from mog_commons.io import print_safe
from mog_commons.types import *
from easy_menu.exceptions import ConfigError, EncodingError

URL_PATTERN = re.compile(r'^http[s]?://')


class Loader(object):
    def __init__(self, work_dir, cache_dir, encoding='utf-8', stdout=sys.stdout, clear_cache=False):
        self.work_dir = work_dir
        self.cache_dir = cache_dir
        self.encoding = encoding
        self.stdout = stdout
        self.clear_cache = clear_cache
        self.cache = {}

    @staticmethod
    def is_url(path):
        return path is not None and bool(URL_PATTERN.match(path))

    @types(is_command=bool, path_or_url_or_cmdline=String, eval_expire=Option(int))
    def load(self, is_command, path_or_url_or_cmdline, eval_expire=None):
        """
        Load data from one file, url or command line, then store to a dict.

        :param is_command: True if using command line output
        :param path_or_url_or_cmdline:
        :param eval_expire: seconds to read
        :return: dict representation of data
        """
        assert path_or_url_or_cmdline

        # normalize file path
        is_url = self.is_url(path_or_url_or_cmdline)
        if not is_command and not is_url:
            if self.work_dir is not None and not os.path.isabs(path_or_url_or_cmdline):
                path_or_url_or_cmdline = os.path.join(self.work_dir, path_or_url_or_cmdline)

        # if already loaded, use cache data
        c = self.cache.get((is_command, path_or_url_or_cmdline))
        if c is not None:
            return c

        try:
            if is_command:
                # execute command
                if eval_expire is None:
                    data = self._eval_command(path_or_url_or_cmdline)
                else:
                    # if eval_expire is defined, check the cache on disk
                    data = self._eval_command_with_cache(path_or_url_or_cmdline, eval_expire)
            elif self.is_url(path_or_url_or_cmdline):
                # read from URL
                print_safe('Reading from URL: %s' % path_or_url_or_cmdline, self.encoding, output=self.stdout)
                data = urlopen(path_or_url_or_cmdline).read()
            else:
                # read from file as bytes
                print_safe('Reading file: %s' % path_or_url_or_cmdline, self.encoding, output=self.stdout)
                with open(path_or_url_or_cmdline, 'rb') as f:
                    data = f.read()

            # decode string with fallback
            data_str = unicode_decode(data, [self.encoding, 'utf-8'])

            # apply jinja2 template rendering
            try:
                data_str = Environment().from_string(data_str).render()
            finally:
                pass

            # load as YAML string
            menu = yaml.load(data_str)

            # update cache data (Note: cache property is mutable!)
            self.cache[(is_command, path_or_url_or_cmdline)] = menu
        except IOError:
            raise ConfigError(path_or_url_or_cmdline, 'Failed to open.')
        except UnicodeDecodeError:
            raise EncodingError('Failed to decode with %s: %s' % (self.encoding, path_or_url_or_cmdline))
        except yaml.YAMLError as e:
            raise ConfigError(path_or_url_or_cmdline, 'YAML format error: %s' % to_unicode(str(e)))
        return menu

    def _eval_command(self, cmdline):
        """
        :param cmdline:
        :return: encoded binary
        """

        # return code and stderr are ignored
        print_safe('Executing: %s' % cmdline, self.encoding, output=self.stdout)
        return capture_command(cmdline, shell=True, cwd=self.work_dir, cmd_encoding=self.encoding)[1]

    @types(cmdline=String, expire=int)
    def _eval_command_with_cache(self, cmdline, expire):
        path = self._eval_cache_path(cmdline)

        # read cache
        is_readable = not self.clear_cache and os.path.exists(path) and (time.time() - os.path.getmtime(path)) < expire
        if is_readable:
            print_safe('Reading eval cache: %s' % path, self.encoding, output=self.stdout)
            with open(path, 'rb') as f:
                return f.read()

        # execute command
        data = self._eval_command(cmdline)

        # write cache
        print_safe('Writing eval cache: %s' % path, self.encoding, output=self.stdout)

        # make parent directory
        if not os.path.exists(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))

        with open(self._eval_cache_path(cmdline), 'wb') as f:
            f.write(data)
        return data

    def _eval_cache_path(self, cmdline):
        h = hashlib.md5(to_bytes(cmdline, self.encoding)).hexdigest()
        return os.path.join(self.cache_dir, h[:2], h[2:])



from __future__ import division, print_function, absolute_import, unicode_literals

import six
from mog_commons.case_class import CaseClass
from mog_commons.string import to_unicode
from mog_commons.collection import get_single_item
from mog_commons.types import *
from easy_menu.entity import Item, Meta
from easy_menu.entity.command_line import CommandLine


class Command(Item):
    @types(title=Unicode, command_lines=ListOf(CommandLine))
    def __init__(self, title, command_lines):
        """
        :param title:
        :param command_lines:
        :return:
        """
        CaseClass.__init__(self, ('title', title), ('command_lines', command_lines))

    @staticmethod
    @types(Item, data=dict, meta=Meta)
    def parse(data, meta, loader, encoding='utf-8', depth=0):
        """
        Parse one command operation.
        :param data: dict:
        :param meta: Meta: meta configuration inherited from the parent menu
        :param loader: not used
        :param encoding: string:
        :param depth: not used
        :return: Command:
        """
        if len(data) != 1:
            raise ValueError('Command should have only one element, not %s.' % len(data))

        title, content = get_single_item(data)
        assert isinstance(title, six.string_types), 'Command title must be string, not %s' % type(title).__name__
        title = to_unicode(title, encoding)

        if isinstance(content, six.string_types):
            # single command
            return Command(title, [CommandLine.parse(content, meta, encoding)])
        elif isinstance(content, list):
            # command list
            return Command(title, [CommandLine.parse(d, meta, encoding) for d in content])
        else:
            raise ValueError('Invalid command content type: %s' % type(content).__name__)

    @types(Unicode)
    def formatted(self):
        return '\n'.join(
            ['* %s:' % self.title] + ['  %s' % line for x in self.command_lines for line in x.formatted().splitlines()])



KEYWORD_META = 'meta'
KEYWORD_INCLUDE = 'include'
KEYWORD_EVAL = 'eval'

from .item import Item
from .meta import Meta
from .command_line import CommandLine
from .command import Command
from .menu import Menu



from __future__ import division, print_function, absolute_import, unicode_literals

import sys
import signal
import os
import socket
import getpass
from mog_commons.terminal import TerminalHandler
from easy_menu.view import Terminal
from easy_menu.controller import CommandExecutor
from easy_menu.setting.setting import Setting
from easy_menu.logger import SystemLogger
from easy_menu.exceptions import EasyMenuError


def get_hostname():
    hostname = socket.gethostname()
    nickname = os.environ.get('NICKNAME')
    return hostname + (' (%s)' % nickname if nickname else '')


def get_username():
    return getpass.getuser()


def main(stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr, keep_input_clean=True, timing=True):
    """
    Main function
    """

    base_setting = Setting(stdin=stdin, stdout=stdout, stderr=stderr).parse_args(sys.argv)

    # for terminal restoration
    handler = TerminalHandler(stdin=stdin, stdout=stdout, stderr=stderr,
                              keep_input_clean=keep_input_clean, getch_enabled=base_setting.getch_enabled)
    signal.signal(signal.SIGTERM, handler.restore_terminal)

    try:
        setting = base_setting.resolve_encoding(handler).lookup_config().load_config()
        executor = CommandExecutor(
            SystemLogger(setting.encoding), setting.encoding, stdin, stdout, stderr, setting.pid_dir
        )

        t = Terminal(
            setting.root_menu,
            get_hostname(),
            get_username(),
            executor,
            handler=handler,
            _input=setting.stdin,
            _output=setting.stdout,
            encoding=setting.encoding,
            lang=setting.lang,
            width=setting.width,
            timing=timing
        )

        t.loop()
    except (KeyboardInterrupt, EOFError):
        pass
    except EasyMenuError as e:
        base_setting.stdout.write('%s: %s\n' % (e.__class__.__name__, e))
        return 2
    except IOError as e:
        # maybe killed by outside
        base_setting.stdout.write('\n%s: %s\n' % (e.__class__.__name__, e))
        return 3
    except OSError as e:
        # e.g. working directory does not exist
        base_setting.stdout.write('%s: %s\n' % (e.__class__.__name__, e))
        return 4
    finally:
        # assume to restore original terminal settings
        handler.restore_terminal(None, None)
    return 0



import sys
from setuptools import setup, find_packages

SRC_DIR = 'src'


def get_version():
    sys.path[:0] = [SRC_DIR]
    return __import__('easy_menu').__version__


setup(
    name='easy-menu',
    version=get_version(),
    description='Super Simple Terminal Command Launcher Generator',
    author='mogproject',
    author_email='mogproj@gmail.com',
    license='Apache 2.0 License',
    url='https://github.com/mogproject/easy-menu',
    install_requires=[
        'pyyaml',
        'six',
        'jinja2' + (' == 2.6' if sys.version_info[:2] == (3, 2) else ''),
        'mog-commons >= 0.2.0',
    ],
    tests_require=[
        'mock == 1.0.1',  # lock version for older version of setuptools
    ] + (['unittest2'] if sys.version_info < (2, 7) else []),
    package_dir={'': SRC_DIR},
    packages=find_packages(SRC_DIR),
    include_package_data=True,
    test_suite='tests',
    entry_points="""
    [console_scripts]
    easy-menu = easy_menu.easy_menu:main
    """,
)



from __future__ import division, print_function, absolute_import, unicode_literals

import os
from mog_commons.command import execute_command_with_pid, pid_exists
from mog_commons.types import *
from easy_menu.entity.command import Command, CommandLine
from easy_menu.logger.logger import Logger
from easy_menu.controller.executor import Executor


class CommandExecutor(Executor):
    @types(logger=Logger)
    def __init__(self, logger, encoding, stdin, stdout, stderr, pid_dir):
        self.logger = logger
        self.encoding = encoding
        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr
        self.pid_dir = pid_dir

    @types(int, command=Command)
    def execute(self, command):
        ret_code = 0
        for command_line in command.command_lines:
            try:
                self.logger.info('Command started: %s' % command_line.cmd)
                if command_line.meta.lock:
                    pid_file = self._pid_file_path(command_line)

                    # create pid directory
                    if not os.path.exists(os.path.dirname(pid_file)):
                        os.makedirs(os.path.dirname(pid_file))
                else:
                    pid_file = None

                ret_code = execute_command_with_pid(
                    command_line.cmd, pid_file=pid_file, shell=True, cwd=command_line.meta.work_dir,
                    env=command_line.meta.env, stdin=self.stdin, stdout=self.stdout, stderr=self.stderr,
                    cmd_encoding=self.encoding)
                self.logger.info('Command ended with return code: %d' % ret_code)

            except KeyboardInterrupt:
                self.logger.info('Command interrupted.')
                ret_code = 130

            # if a command fails, the successors will not run
            if ret_code != 0:
                break
        return ret_code

    @types(bool, command=Command)
    def is_running(self, command):
        def check_pid(cmdline):
            if not cmdline.meta.lock:
                return False

            path = self._pid_file_path(cmdline)
            if not os.path.exists(path):
                return False

            with open(path, 'r') as f:
                pid = int(f.read())
            return pid_exists(pid)

        return any(check_pid(c) for c in command.command_lines)

    @types(String, cmdline=CommandLine)
    def _pid_file_path(self, cmdline):
        h = cmdline.to_hash_string()
        return os.path.join(self.pid_dir, h[:2], h[2:])
